% :- use_module(library(occurs), [contains_var/2]).
%
%   contains_var(+Variable, +Term)
%   is true when the given Term contains at least one sub-term which
%   is identical to the given Variable.  We use '=='/2 to check for
%   the variable (contains_term/2 uses '=') so it can be used to check
%   for arbitrary terms, not just variables.

contains_var(Variable, Term) :-
        \+ free_of_var(Variable, Term).

%   free_of_var(+Variable, +Term)
%   is true when the given Term contains no sub-term identical to the
%   given Variable (which may actually be any term, not just a var).
%   For variables, this is precisely the "occurs check" which is
%   needed for sound unification.

free_of_var(Variable, Term) :-
        Term == Variable,
        !,
        fail.
free_of_var(Variable, Term) :-
        compound(Term),
        !,
        functor(Term, _, Arity),
        free_of_var(Arity, Term, Variable).
free_of_var(_, _).

free_of_var(1, Term, Variable) :- !,
        arg(1, Term, Argument),
        free_of_var(Variable, Argument).
free_of_var(N, Term, Variable) :-
        arg(N, Term, Argument),
        free_of_var(Variable, Argument),
        M is N-1, !,
        free_of_var(M, Term, Variable).

% :- use_module(library(arg), [genarg/3]).
%
genarg(N, Term, Arg) :-
        integer(N),
        nonvar(Term),
        !,
        arg(N, Term, Arg).
genarg(N, Term, Arg) :-
        var(N),
        nonvar(Term),
        !,
        functor(Term, _, Arity),
        genarg(Arity, Term, Arg, N).

genarg(1, Term, Arg, 1) :- !,
        arg(1, Term, Arg).
genarg(N, Term, Arg, N) :-
        arg(N, Term, Arg).
genarg(K, Term, Arg, N) :-
        K > 1, J is K-1,
        genarg(J, Term, Arg, N).

%------------------------------------------------------------------------------
% greater( S, T ):
%	S is greater than T with respect to the lexicographic path ordering
%------------------------------------------------------------------------------

greater( S, _ ):-
	var(S),
	!,
	fail.
greater( S, T ):-
	\+var(T),
	!,
	functor(S,F,N),
	functor(T,G,M),
	greater(F,G,N,M,S,T).
greater( S, T ):-
	% T is a variable
	functor(S,_F,_),
	contains_var(T,S),
	!.
	

greater(F,G,N,M,S,T):-
	equivalent_op(F,N,G,M),
	!,
        lexicographic(0,S,N,T,M)
	;
	( genarg(_,S,S1),
	  greaterEqual( S1, T)
	), 
	!.
greater(F,G,N,M,S,T):-
	(greater_op(F,N,G,M)->
	    greater_than_list(S,0,T,M)
	; 
	    genarg(_,S,S1),
	    greaterEqual(S1, T)
	), 
	!.
	
lexicographic(N,_,N,_,_M):-
	!,
	fail.
lexicographic(M,_,_N,_,M):-
	!,
	fail.
lexicographic(I,S,N,T,M):-
	I1 is I+1,
	arg(I1,S,SI),
	arg(I1,T,TI),
	(SI==TI ->
	     lexicographic(I1,S,N,T,M)
	;greater(SI,TI) ->
	     greater_than_list(S,I1,T,M)).

greater_than_list(_,N,_,N):-
	!.
greater_than_list(S,I,T,M):-
	I1 is I+1,
	arg(I1,T,TI),
	greater(S,TI),
	greater_than_list(S,I1,T,M).

%------------------------------------------------------------------------------
% greaterEqual( S, T ):
%	S is greater than T with respect lpo or S is equal to T
%------------------------------------------------------------------------------

greaterEqual(T1,T2):- 
	T1 == T2,
	!.
greaterEqual(T1,T2):- 
	greater(T1,T2),
	!.

%------------------------------------------------------------------------------
% maximal_term( S, T, M ):
%------------------------------------------------------------------------------

maximal_term( S, T, S ):- greaterEqual( S, T ), !.
maximal_term( S, T, T ):- greater( T, S ), !.


equivalent_op(F,N,F,N) :-
	!.
equivalent_op(F,N,G,N) :-
	option(mode,saturate),
	isSkolemFunction(F),
	isSkolemFunction(G),
	!.
equivalent_op(F,N,G,N) :-
	option(mode,saturate),
	\+ isSkolemFunction(F),
	\+ isSkolemFunction(G),
	!.

greater_op(F,N,G,N) :-
	isSkolemFunction(F),
	isSkolemFunction(G),
	!,
	F @> G.
greater_op(F,_N,G,_M) :-
	isSkolemFunction(F),
	!,
	\+ isSkolemFunction(G).
greater_op(F,N,G,M) :-
	\+ isSkolemFunction(F),
	\+ isSkolemFunction(G),
	N > M.
greater_op(F,N,G,N) :-
	option(mode,satisfy),
	\+ isSkolemFunction(G),
	F @> G.

isSkolemFunction(F) :-
	name(F,[102|L]),
	numberChars(L),
	!.

numberChars([]).
numberChars([N|L]) :-
	N >= 48,
	N =< 57,
	numberChars(L).